import base64
import os

from odoo.addons.saas_worker.tests.common import SaasWorkerHttpCommon, _hss
from odoo.addons.saas_worker.auth_util import SaasAuth
from odoo.addons.saas_master.saas_master import DOMAIN_SAAS


class TestSaasMasterController(SaasWorkerHttpCommon):
    @classmethod
    def setUpClass(cls):
        super().setUpClass()

        # Add 'build' host
        cls._mock_default_saas_worker_list = list(cls._mock_default_saas_worker_list) + [
            {
                "name": "builder.odoo.localhost",
                "public_ip": "127.42.6.1",
                "mode": "build",
                "http_shared_secret": _hss("builder-1"),
            },
        ]

        # Create tests datacenters, servers, ...
        cls.Dc = cls.env["saas.dc"]
        cls.Server = cls.env["saas.server"]
        cls.gra2 = cls.Dc.create({"provider": "OVH", "name": "GRA2"})
        cls.rbx8 = cls.Dc.create({"provider": "OVH", "name": "RBX8"})
        cls.euw1b = cls.Dc.create({"provider": "Google", "name": "EUROPE-WEST1-B"})
        cls.euw3a = cls.Dc.create({"provider": "Google", "name": "EUROPE-WEST3-A"})
        cls.usc1a = cls.Dc.create({"provider": "Google", "name": "US-CENTRAL1-A"})
        cls.use1b = cls.Dc.create({"provider": "Google", "name": "US-EAST1-B"})
        cls.usw1c = cls.Dc.create({"provider": "Google", "name": "US-WEST1-C"})

        # EU trials
        cls.eu01a = cls.Server.create(
            {
                "name": f"eu01a.{DOMAIN_SAAS}",
                "dc_id": cls.rbx8.id,
                "backuped": False,
                "hosting": "eu",
                "public_ip": "10.10.10.10",
                "mode": "trial",
            }
        )
        cls.eu01b = cls.Server.create(
            {
                "name": f"eu01b.{DOMAIN_SAAS}",
                "dc_id": cls.gra2.id,
                "backuped": False,
                "hosting": "eu",
                "public_ip": "10.10.10.11",
                "mode": "trial slave",
                "master": f"{cls.eu01a.name}.{DOMAIN_SAAS}",
                "primaries": [(6, 0, [cls.eu01a.id])],
            }
        )
        cls.eu01c = cls.Server.create(
            {
                "name": f"adv2-001.{DOMAIN_SAAS}",
                "dc_id": cls.gra2.id,
                "backuped": False,
                "hosting": "eu",
                "public_ip": "10.10.10.12",
                "mode": "trial slave",
            }
        )
        # EU standbys
        cls.standby_eu01 = cls.Server.create(
            {
                "name": f"standby-eu01.{DOMAIN_SAAS}",
                "dc_id": cls.euw1b.id,
                "backuped": False,
                "hosting": "eu",
                "public_ip": "10.10.20.10",
                "mode": "pg_standby",
            }
        )
        cls.standby_eu02 = cls.Server.create(
            {
                "name": f"standby-eu02.{DOMAIN_SAAS}",
                "dc_id": cls.euw3a.id,
                "backuped": False,
                "hosting": "eu",
                "public_ip": "10.10.20.11",
                "mode": "pg_standby",
                "primaries": [(6, 0, [cls.eu01a.id])],
            }
        )
        # AM trials
        cls.am01a = cls.Server.create(
            {
                "name": f"am01a.{DOMAIN_SAAS}",
                "dc_id": cls.usc1a.id,
                "backuped": False,
                "hosting": "am",
                "public_ip": "10.10.10.20",
                "mode": "trial",
            }
        )

        # AM standbys
        cls.standby_am01 = cls.Server.create(
            {
                "name": f"standby-am01.{DOMAIN_SAAS}",
                "dc_id": cls.use1b.id,
                "backuped": False,
                "hosting": "am",
                "public_ip": "10.10.20.20",
                "mode": "pg_standby",
                "primaries": [(6, 0, [cls.am01a.id])],
            }
        )
        cls.standby_am02 = cls.Server.create(
            {
                "name": f"standby-am02.{DOMAIN_SAAS}",
                "dc_id": cls.usw1c.id,
                "backuped": False,
                "hosting": "am",
                "public_ip": "10.10.20.21",
                "mode": "pg_standby",
            }
        )

    def setUp(self):
        super().setUp()

        # baseline
        self.assertEqual(len(self.am01a.replicas), 1)
        self.assertEqual(len(self.standby_am01.primaries), 1)
        self.assertEqual(len(self.standby_am02.primaries), 0)
        self.assertEqual(len(self.eu01a.replicas), 2)
        self.assertEqual(len(self.eu01b.primaries), 1)
        self.assertEqual(len(self.eu01c.primaries), 0)
        self.assertEqual(len(self.eu01c.replicas), 0)
        self.assertEqual(len(self.standby_eu01.primaries), 0)
        self.assertEqual(len(self.standby_eu02.primaries), 1)

    def test_server_create_without_replication(self):
        """
        Creating am02a without any replication must:
        - create a new server record
        - not set any of the replication-related fields
        """
        hostname = "am02a"
        ip = "10.10.10.21"
        private_ipv4 = "10.0.0.42"
        mode = "trial new"
        hosting = "am"
        http_secret = str(base64.b64encode(os.urandom(32)))
        machine_type = "c3d-standard-16"

        with self._mock_saas_config(from_role="build"):
            resp = self.url_open(
                "/saas_master/server/create",
                auth=SaasAuth(),
                json={
                    "params": {
                        "hostname": hostname,
                        "public_ip": ip,
                        "mode": mode,
                        "hosting": hosting,
                        "data_center": self.usc1a.name,
                        "http_shared_secret": http_secret,
                        "disable_backup": True,
                        "private_ipv4": private_ipv4,
                        "machine_type": machine_type,
                    }
                },
            )
            resp.raise_for_status()
            res = resp.json().get("result", {})

        am02a = self.env["saas.server"].browse(res["server_id"])
        self.assertEqual(am02a.dc_id, self.usc1a)
        self.assertEqual(am02a.backuped, False)
        self.assertEqual(am02a.name, f"{hostname}.{DOMAIN_SAAS}")
        self.assertEqual(am02a.hosting, hosting)
        self.assertEqual(am02a.http_shared_secret, http_secret)
        self.assertEqual(am02a.public_ip, ip)
        self.assertEqual(am02a.private_ip, private_ipv4)
        self.assertEqual(am02a.mode, mode)
        self.assertFalse(am02a.master)
        self.assertFalse(am02a.primaries)
        self.assertFalse(am02a.replicas)
        self.assertEqual(am02a.machine_family, "C3D")

    def test_server_create_with_dedicated_replication(self):
        """
        Creating am01b to be a dedicated replica of am01a must:
        - set the master field on am01b
        - add am01b to am01a's replicas
        - add am01a to am01b's primaries
        """
        hostname = "am01b"
        ip = "10.10.10.21"
        mode = "trial new"
        hosting = "am"
        http_secret = str(base64.b64encode(os.urandom(32)))

        for replication_mode in ("dedicated", "hybrid"):
            with self.subTest(replication_mode):
                with self._mock_saas_config(from_role="build"):
                    resp = self.url_open(
                        "/saas_master/server/create",
                        auth=SaasAuth(),
                        json={
                            "params": {
                                "hostname": hostname,
                                "public_ip": ip,
                                "mode": mode,
                                "hosting": hosting,
                                "data_center": self.usc1a.name,
                                "http_shared_secret": http_secret,
                                "disable_backup": True,
                                "master": self.am01a.name,
                                "replication_mode": replication_mode,
                            }
                        },
                    )
                    resp.raise_for_status()
                    res = resp.json().get("result", {})

                am01b = self.env["saas.server"].browse(res["server_id"])
                self.assertEqual(am01b.master, self.am01a.name)
                self.assertIn(self.am01a, am01b.primaries)
                self.assertIn(am01b, self.am01a.replicas)
                self.assertEqual(len(self.am01a.replicas), 2)
                self.assertEqual(len(am01b.primaries), 1)
                am01b.unlink()

    def test_server_create_with_shared_replication(self):
        """
        Creating am02a to have replicas standby-am01 and standby-am02 must:
        - add standby-am01 and standby-am02 to am02a's replicas
        - add am02a to standby-am01's primaries
        - add am02a to standby-am02's primaries
        """

        for replication_mode in ("shared", "hybrid"):
            with self.subTest(replication_mode):
                with self._mock_saas_config(from_role="build"):
                    resp = self.url_open(
                        "/saas_master/server/create",
                        auth=SaasAuth(),
                        json={
                            "params": {
                                "hostname": "am02a",
                                "public_ip": "10.10.10.21",
                                "mode": "trial",
                                "hosting": "eu",
                                "data_center": self.gra2.name,
                                "replicas": ["standby-am01", "standby-am02"],
                                "disable_backup": True,
                                "replication_mode": replication_mode,
                            }
                        },
                    )
                    resp.raise_for_status()
                    res = resp.json().get("result", {})

                server = self.env["saas.server"].browse(res["server_id"])
                self.assertTrue(server)
                self.assertEqual(len(server.primaries), 0)
                self.assertEqual(len(server.replicas), 2)
                self.assertIn(self.standby_am01, server.replicas)
                self.assertIn(self.standby_am02, server.replicas)
                self.assertEqual(
                    len(self.standby_am01.primaries), 2, "shared replication must add a new server"
                )
                self.assertEqual(len(self.standby_am02.primaries), 1)
                server.unlink()

    def test_server_assign_without_replication(self):
        """
        Assigning eu01a without any replication must:
        - reset the master field on eu01a
        - reset the primaries field on eu01a
        - reset the replicas field on eu01a
        """

        with self._mock_saas_config(from_role="build"):
            resp = self.url_open(
                "/saas_master/server/assign",
                auth=SaasAuth(),
                json={
                    "params": {
                        "hostname": "eu01a",
                        "public_ip": self.eu01a.public_ip,
                        "mode": self.eu01a.mode,
                        "master": self.eu01a.name,
                        "disable_backup": True,
                    }
                },
            )
            resp.raise_for_status()
            res = resp.json().get("result", {})

        eu01a = self.env["saas.server"].browse(res["server_id"])
        self.assertFalse(eu01a.master)
        self.assertFalse(eu01a.primaries)
        self.assertFalse(eu01a.replicas)

    def test_server_assign_with_dedicated_replication(self):
        """
        Assigning eu01c to be a dedicated replica of eu01a must:
        - set the master field on eu01c
        - add eu01c to eu01a's replicas
        - add eu01a to eu01c's primaries
        """

        for replication_mode in ("dedicated", "hybrid"):
            with self.subTest(replication_mode):
                with self._mock_saas_config(from_role="build"):
                    resp = self.url_open(
                        "/saas_master/server/assign",
                        auth=SaasAuth(),
                        json={
                            "params": {
                                "hostname": "eu01c",
                                "public_ip": self.eu01c.public_ip,
                                "mode": self.eu01c.mode,
                                "master": self.eu01a.name,
                                "disable_backup": True,
                                "replication_mode": replication_mode,
                            }
                        },
                    )
                    resp.raise_for_status()
                    res = resp.json().get("result", {})

                eu01c = self.env["saas.server"].browse(res["server_id"])
                self.assertTrue(eu01c)
                self.assertEqual(self.eu01c, eu01c)
                self.assertEqual(eu01c.master, self.eu01a.name)
                self.assertIn(self.eu01a, eu01c.primaries)
                self.assertIn(eu01c, self.eu01a.replicas)
                self.assertEqual(len(self.eu01a.replicas), 3)
                self.assertEqual(len(self.eu01b.primaries), 1)
                self.assertEqual(len(eu01c.primaries), 1)

    def test_server_assign_with_shared_replication(self):
        """
        Assigning eu01c to have replicas standby-eu01 and standby-eu02 must:
        - add standby-eu01 and standby-eu02 to eu01c's replicas
        - add eu01c to standby-eu01's primaries
        - add eu01c to standby-eu02's primaries
        """
        for replication_mode in ("shared", "hybrid"):
            with self.subTest(replication_mode):
                with self._mock_saas_config(from_role="build"):
                    resp = self.url_open(
                        "/saas_master/server/assign",
                        auth=SaasAuth(),
                        json={
                            "params": {
                                "hostname": "eu01c",
                                "public_ip": self.eu01c.public_ip,
                                "mode": "trial",
                                "replicas": ["standby-eu01", "standby-eu02"],
                                "disable_backup": True,
                                "replication_mode": replication_mode,
                            }
                        },
                    )
                    resp.raise_for_status()
                    res = resp.json().get("result", {})

                eu01c = self.env["saas.server"].browse(res["server_id"])
                self.assertTrue(eu01c)
                self.assertEqual(self.eu01c, eu01c)
                self.assertFalse(eu01c.master)  # used for dedicated replication only
                self.assertIn(self.standby_eu01, eu01c.replicas)
                self.assertIn(self.standby_eu02, eu01c.replicas)
                self.assertIn(eu01c, self.standby_eu01.primaries)
                self.assertIn(eu01c, self.standby_eu02.primaries)
                self.assertEqual(len(self.eu01c.replicas), 2)
                self.assertEqual(len(self.standby_eu01.primaries), 1)
                self.assertEqual(len(self.standby_eu02.primaries), 2)
